from magma import *
from mantle import Mux
from .register import _RegisterName, Register

__all__ = ['DefinePIPO', 'PIPO']

def DefinePIPO(n, init=0, has_ce=False, has_reset=False):
    """
    Generate Parallel-In, Parallel-Out shift register.

    SI : In(Bit), PI : In(Bits(n)), LOAD : Bit, O : Out(Bits(n))
    """

    T = Bits(n)
    class _PIPO(Circuit):
        name = _RegisterName('PIPO', n, init, has_ce, has_reset)
        IO = ['SI', In(Bit), 'PI', In(T), 'LOAD', In(Bit),
              'O', Out(T)] + ClockInterface(has_ce,has_reset)
        @classmethod
        def definition(pipo):
            def mux2(y):
                return curry(Mux(2), prefix='I')

            mux = braid(col(mux2, n), forkargs=['S'])
            reg = Register(n, init=init, has_ce=has_ce, has_reset=has_reset)

            #si = array([pipo.SI] + [reg.O[i] for i in range(n-1)])
            si = concat(array(pipo.SI),reg.O[0:n-1])
            mux(si, pipo.PI, pipo.LOAD)
            reg(mux)
            wire(reg.O, pipo.O)
            wireclock(pipo, reg)

    return _PIPO

def PIPO(n, init=0, has_ce=False, has_reset=False, **kwargs):
    return DefinePIPO(n, init, has_ce, has_reset)(**kwargs)




import os
from magma import *
from mantle import *
from mantle.common.sipo import SIPO

__all__ = ['DefineLFSR', '_lfsrtaps']

_lfsrtaps = {}


def DefineLFSR(n, init=1, has_ce=False, has_reset=False):
    def readtaps():
        global _lfsrtaps

        if len(_lfsrtaps) != 0:
            return

        import csv

        name = os.path.join(os.path.dirname(__file__), 'lfsr.csv')

        f = open(name)
        row = csv.reader(f)

        for data in row:
            if data[0].isdigit():
                tap = int(data[0])
                _lfsrtaps[tap] = [int(v) for v in data[1].split(',')]

    readtaps()

    tap = _lfsrtaps[n]
    nt = len(tap)

    name = 'LFSR{}'.format(n)
    if init:
        name += '_' + str(init)
    if has_ce:
        name += 'CE'
    if has_reset:
        name += 'R'
    lfsr = DefineCircuit(name, "O", Out(Bits(n)), 
             *ClockInterface(has_ce=has_ce, has_reset=has_reset))
    shift = SIPO(n, init=init, has_ce=has_ce, has_reset=has_reset)

    t = []
    for i in range(nt):
        t.append(shift.O[tap[i] - 1])
    t = array(t)

    s = uncurry(XOr(nt))(t)
    shift(s)

    wire(shift.O, lfsr.O)
    wire(shift.CLK, lfsr.CLK)
    if has_reset:
        wire(shift.RESET, lfsr.RESET)
    if has_ce:
        wire(shift.CE, lfsr.CE)
    EndDefine()
    return lfsr



from ..ice40.RAMB import RAMB, ROMB

__all__ = ['Memory']

#
# Contruct a memory mem[height][width]
#
def Memory(height, width, rom=None, readonly=False):

    if rom:
        assert height == len(rom)
    else:
        rom = height * [0]

    return ROMB(height, width, rom) if readonly else RAMB(height, width, rom)



from magma import *
from .swap import swap, EvenOddSwaps
from .permute import Riffle, UnRiffle

__all__  = ['DefineEvenOddMerger', 'EvenOddMerger', 'evenoddmerger']
__all__  = ['DefineEvenOddSorter', 'EvenOddSorter', 'evenoddsorter']

#
# Convert an unsorted sequence into a sorted sequence
#
# EvenOdd(n) => 2 EvenOddMerger(n//2) => EvenOddSwaps(n)) 
#
def DefineEvenOddMerger(n):
    assert n in [2, 4, 8, 16]
    T = Bits(n)
    class _EvenOddMerger(Circuit):
        name = 'EvenOddMerger{}'.format(n)
        IO = ['I', In(T), "O", Out(T)]
        @classmethod
        def definition(io):
            if n == 2: # Sort 2 element sequences
                wire(swap(io.I), io.O)
            else:
                unriffle = UnRiffle(n)
                merger0 = EvenOddMerger(n//2) # bot merger 
                merger1 = EvenOddMerger(n//2) # top merger
                merger = flat( join(merger0, merger1) )
                riffle = Riffle(n)
                merger =  compose( riffle, compose( merger, unriffle ) )
                evenoddswap = EvenOddSwaps(n)
                merger = compose( evenoddswap, merger )
                wire( merger(io.I), io.O )
    return _EvenOddMerger

def EvenOddMerger(n):
    return DefineEvenOddMerger(n)()

def evenoddmerger(I):
    return EvenOddMerger(len(I))(I)

#
# Convert an unsorted sequence into a sorted sequence
#
# 2 EvenOddSorter(n//2) => EvenOddMerger(n)) 
#
def DefineEvenOddSorter(n):
    assert n in [2, 4, 8, 16]
    T = Bits(n)
    class _EvenOddSorter(Circuit):
        name = 'EvenOddSorter{}'.format(n)
        IO = ['I', In(T), "O", Out(T)]
        @classmethod
        def definition(io):
            if n == 2: # Sort 2 element sequences
                wire(swap(io.I), io.O)
            else:
                sorter0 = EvenOddSorter(n//2) # bot sorter 
                sorter1 = EvenOddSorter(n//2) # top sorter
                sorter = flat( join(sorter0, sorter1) )
                merger = EvenOddMerger(n)
                sorter = compose( merger, sorter )
                wire( sorter(io.I), io.O )
    return _EvenOddSorter

def EvenOddSorter(n):
    return DefineEvenOddSorter(n)()

def evenoddsorter(I):
    return EvenOddSorter(len(I))(I)




from magma import Circuit, Bits, In, Out, wire, join
from .LUT import LUT1, A0

__all__ = ['DefineBuf', 'Buf']

def DefineBuf(width=1):
    """
    Generate Buf module

    I : In(Bits(width)), O : Out(Bits(width))
    """
    T = Bits(width)
    class BufN(Circuit):

        name = 'Buf{}'.format(width)
        IO  = ['I', In(T), 'O', Out(T)]

        @classmethod
        def definition(def_):
            def buf(y):
                return Buf(loc=(0,y//8, y%8))
            buffer = join(col(buf, width))
            wire(def_.I, buffer.I0)
            wire(buffer.O, def_.O)

    return BufN

def Buf(width=None, **kwargs):
    if width is None:
        return LUT1(A0, **kwargs)
    return DefineBuf(width)(**kwargs)




import six
from magma.bitutils import *
from collections import defaultdict

def simulate_sb_carry(self, value_store, state_store):
    I0 = value_store.get_value(self.I0)
    I1 = value_store.get_value(self.I1)
    CI = value_store.get_value(self.CI)
    val = (I0 and I1) or (I1 and CI) or (I0 and CI)
    value_store.set_value(self.CO, val)

def simulate_sb_lut4(self, value_store, state_store):
    init = self.kwargs['LUT_INIT']
    if isinstance(init, six.string_types):
        a = init.split("'")
        l = int(a[0])
        p = a[1][0]
        if p == 'h':
            base = 16
        else:
            base = 10

        i = int(a[1][1:], base)
        init = lutinit(i, l)

    if isinstance(init, int):
        init = (init, 16)
    lut = int2seq(init[0], init[1])
    lut = [bool(i) for i in lut]
    inputs = self.interface.inputs()
    I = [value_store.get_value(inputs[x]) for x in range(0, 4)]
    idx = seq2int(I)
    val = lut[idx]
    value_store.set_value(self.O, val)

# SB_DFF*
# ce = Clock Enable
# r = Reset
# sy = Synchronous
# s = Set
# n = Negative Edge
def gen_sb_dff_sim(ce=False, sy=False, r=False, s=False, n=False):
    def sim(self, value_store, state_store):
        cur_clock = value_store.get_value(self.C)

        if not state_store:
            state_store['prev_clock'] = cur_clock
            state_store['cur_val'] = False

        if r:
            cur_r = value_store.get_value(self.R)
        if s:
            cur_s = value_store.get_value(self.S)

        prev_clock = state_store['prev_clock']
        if not n:
            clock_edge = cur_clock and not prev_clock
        else:
            clock_edge = not cur_clock and prev_clock

        new_val = state_store['cur_val']

        if clock_edge:
            input_val = value_store.get_value(self.D)

            enable = True
            if ce:
                enable = value_store.get_value(self.E)

            if enable:
                if r and sy and cur_r:
                    new_val = False
                elif s and sy and cur_s:
                    new_val = True
                else:
                    new_val = input_val

        if r and not sy and cur_r:
            new_val = False
        if s and not sy and cur_s:
            new_val = True

        state_store['prev_clock'] = cur_clock
        state_store['cur_val'] = new_val
        value_store.set_value(self.Q, new_val)

    return sim

# RAM40_4K*
# prc = Positive Edge Read Clock, pwc = Positive Edge Write Clock
def gen_sb_ram40_4k_sim(prc=True, pwc=True):
    def sim(self, value_store, state_store):
        r_clock = value_store.get_value(self.RCLK)
        w_clock = value_store.get_value(self.WCLK)

        w_mode = int(self.kwargs['WRITE_MODE'])
        r_mode = int(self.kwargs['READ_MODE'])
        n_w_blocks = 16 // pow(2, w_mode) # number of blocks the RAM is split into for writing
        n_r_blocks = 16 // pow(2, r_mode) # number of blocks the RAM is split into for reading
        w_block_sz = 7 + w_mode        # how many bits are req to convert address (7 is for 8-bit),
        r_block_sz = 7 + r_mode        # (8 is for 9-bit), etc.

        # initialize RAM
        if not state_store:
            # block
            state_store['block'] = [False] * 4096
            # The SB_RAM primitive has a special initialization mapping for bits,
            # see the init function in mantle/lattice/ice40/RAMB.py
            # Here we invert the mapping and store the memory in a cartesian
            # grid
            N = n_r_blocks
            M = 16 // N
            for i in range(16):
                v = self.kwargs['INIT_' + str(format(i, 'X'))][0]
                for b in range(256):
                    col = (b//M)%N
                    row = 256*(b%M) + 16*i + b//16
                    state_store['block'][row * N + col] = bool((v >> b) & 1)

            state_store['prev_r_clock'] = r_clock
            state_store['prev_w_clock'] = w_clock

        block = state_store['block']

        # obtain clock information
        prev_w_clock = state_store['prev_w_clock']
        if pwc:
            w_clock_edge = w_clock and not prev_w_clock
        else:
            w_clock_edge = not w_clock and prev_w_clock

        prev_r_clock = state_store['prev_r_clock']
        if prc:
            r_clock_edge = r_clock and not prev_r_clock
        else:
            r_clock_edge = not r_clock and prev_r_clock

        # write to RAM
        if w_clock_edge:
            WDATA = value_store.get_value(self.WDATA)
            w_data = [False for _ in range(n_w_blocks)]
            for i in range(n_r_blocks):
                if n_r_blocks == 16:
                    w_data[i] = WDATA[i]
                elif n_r_blocks == 8:
                    w_data[i] = WDATA[i * 2]
                elif n_r_blocks == 4:
                    w_data[i] = WDATA[1 + i * 4]
                elif n_r_blocks == 2:
                    w_data[i] = WDATA[3 + i * 8]
            mask = value_store.get_value(self.MASK)
            w_addr = value_store.get_value(self.WADDR)
            w_e = value_store.get_value(self.WE)
            w_clke = value_store.get_value(self.WCLKE)
            if w_mode == 0: # use MASK
                if w_e and w_clke:
                    idx = n_w_blocks * seq2int(w_addr)
                    for i in range(n_w_blocks):
                        if mask[i] == False:
                            assert w_data[i] == True or w_data[i] == False
                            block[idx + i] = w_data[i]
            else:
                if w_e and w_clke:
                    idx = n_w_blocks * seq2int(w_addr)
                    for i in range(n_w_blocks):
                        assert w_data[i] == True or w_data[i] == False
                        block[idx + i] = w_data[i]

        # read from RAM
        if r_clock_edge:
            r_addr = value_store.get_value(self.RADDR)

        r_e = value_store.get_value(self.RE)
        r_clke = value_store.get_value(self.RCLKE)

        RDATA = [False for _ in range(16)]
        if r_e and r_clock_edge and r_clke:
            r_data = [False] * n_r_blocks
            row = n_r_blocks * seq2int(r_addr)
            for col in range(n_r_blocks):
                r_data[col] = block[row + col];

            for i in range(n_r_blocks):
                if n_r_blocks == 16:
                    RDATA[i] = r_data[i]
                elif n_r_blocks == 8:
                    RDATA[i * 2] = r_data[i]
                elif n_r_blocks == 4:
                    RDATA[1 + i * 4] = r_data[i]
                elif n_r_blocks == 2:
                    RDATA[3 + i * 8] = r_data[i]

            value_store.set_value(self.RDATA, RDATA)

        state_store['prev_w_clock'] = w_clock
        state_store['prev_r_clock'] = r_clock

    return sim



from magma import Circuit, In, Out, Clock, Bit, Bits, wire, bits, repeat, cache_definition
from mantle import And
from mantle import Mux
from .register import Register
from .decoder import Decoder

__all__  = ['DefineRAM', 'RAM']
__all__ += ['DefineDualRAM', 'DualRAM']

def REGs(n, width, has_ce):
    return [Register(width, has_ce=has_ce) for i in range(n)]

def MUXs(n, width):
    return [Mux(2,width) for i in range(n)]


def readport(height, width, regs, raddr):
    n = 1 << height

    muxs = MUXs(n-1, width)
    for i in range(n//2):
        muxs[i](regs[2*i], regs[2*i+1], raddr[0])

    k = 0
    l = 1 << (height-1)
    for i in range(height-1):
        for j in range(l//2):
            muxs[k+l+j](muxs[k+2*j], muxs[k+2*j+1], raddr[i+1])
        k += l
        l //= 2

    return muxs[n-2]


def writeport(height, width, regs, WADDR, I, WE):
    n = 1 << height

    decoder = Decoder(height)
    enable = And(2,n)
    enable(decoder(WADDR), repeat(WE, n))

    for i in range(n):
        regs[i](I, CE=enable.O[i])


def DefineRAM(height, width):
    n = 1 << height
    TADDR = Bits(height)
    TDATA = Bits(width)

    class _RAM(Circuit):
        name = f'RAM{n}x{width}'
        IO = ['RADDR', In(TADDR),
              'RDATA', Out(TDATA),
              'WADDR', In(TADDR),
              'WDATA', In(TDATA),
              'WE', In(Bit),
              'CLK', In(Clock)
             ]

        @classmethod
        def definition(io):
            regs = REGs(n, width, has_ce=True)
            writeport(height, width, regs, io.WADDR, io.WDATA, io.WE)
            wire( readport(height, width, regs, io.RADDR), io.RDATA )

    return _RAM

def RAM(height, width):
    return DefineRAM(height, width)()


def DefineDualRAM(height, width):
    n = 1 << height
    TADDR = Bits(height)
    TDATA = Bits(width)

    class _DualRAM(Circuit):
        name = f'DualRAM{n}x{width}'
        IO = ['RADDR0', In(TADDR),
              'RDATA0', Out(TDATA),
              'RADDR1', In(TADDR),
              'RDATA1', Out(TDATA),
              'WADDR', In(TADDR),
              'WDATA', In(TDATA),
              'WE', In(Bit),
              'CLK', In(Clock)]

        @classmethod
        def definition(io):
            regs = REGs(n, width, has_ce=True)
            writeport(height, width, regs, io.WADDR, io.WDATA, io.WE)
            wire( readport(height, width, regs, io.RADDR0), io.RDATA0 )
            wire( readport(height, width, regs, io.RADDR1), io.RDATA1 )

    return _DualRAM

def DualRAM(height, width):
    return DefineDualRAM(height, width)()





import math
from magma import *
from mantle import And, Decode, Or
from .counter import Counter, _CounterName

__all__ = ['DefineUpCounterModM', 'UpCounterModM']
__all__ += ['DefineCounterModM', 'CounterModM', 'SizedCounterModM']

#
# Create an n-bit mod-m counter
#
def DefineCounterModM(m, n, cin=False, cout=True, incr=1, 
    has_ce=False, has_reset=False):

    name = _CounterName(f'Counter{n}_Mod{m}', incr, has_ce, has_reset, cin, cout)
    args = []
    if cin:
        args += ['CIN', In(Bit)]

    args += ["O", Out(UInt(n))]
    if cout:
        args += ["COUT", Out(Bit)]

    args += ClockInterface(has_ce, has_reset)

    CounterModM = DefineCircuit(name, *args)

    counter = Counter(n, cin=cin, cout=False, incr=incr, 
                   has_ce=has_ce, has_reset=True)
    reset = Decode(m - 1, n)(counter.O)

    if has_reset:
        reset = Or(2)(reset, CounterModM.RESET)

    if has_ce:
        CE = CounterModM.CE
        reset = And(2)(reset, CE)
        # reset is sometimes called rollover or RO
        # note that we don't return RO in Counter

        # should also handle r in the definition

    wire(reset, counter.RESET) # synchronous reset

    if has_ce:
        wire(CE, counter.CE)

    if cin:
        wire( CounterModM.CIN, counter.CIN )

    wire( counter.O, CounterModM.O )

    if cout:
        wire( reset, CounterModM.COUT )

    EndCircuit()

    return CounterModM

def CounterModM(m, n, cin=False, cout=True, incr=1, 
    has_ce=False, has_reset=False, **kwargs):
    return DefineCounterModM(m, n, cin, cout, incr, has_ce, has_reset=has_reset)(**kwargs)

def SizedCounterModM(m, cin=False, cout=False, incr=1, 
    has_ce=False, **kwargs):
    """
    This is that counts from 0 to m - 1 that uses the minimum number of bits
    :param m: The value the counter counts up to
    :param cin: Whether this counter should have a carry input
    :param cout: Whether this counter should a carry output
    :param incr: How much this counter should increment by per clock. Default is 1.
    :param has_ce: Whether this counter should a clock-enable input
    :param kwargs: Args passed to the counter circuit when it is being initialized
    :return: A counter circuit
    """
    return DefineCounterModM(m, math.ceil(math.log(m, 2)), cin, cout, incr, has_ce)(**kwargs)

DefineUpCounterModM = DefineCounterModM
UpCounterModM = CounterModM



import magma as m
from bit_vector import SIntVector, UIntVector

default_port_mapping = {
    "I": "in",
    "I0": "in0",
    "I1": "in1",
    "O": "out",
    "S": "sel",
}


def DefineCoreirCircuit(*args, **kwargs):
    return m.DefineCircuit(*args, **kwargs,
                           renamed_ports=default_port_mapping)


def DeclareCoreirCircuit(*args, **kwargs):
    return m.DeclareCircuit(*args, **kwargs,
                            renamed_ports=default_port_mapping)


def get_int_vector_type(signed):
    return SIntVector if signed else UIntVector



from magma import *
from mantle import LUT, DefineAdd

__all__ = ['DefineArbiter', 'Arbiter', 'arbiter']

#
# Given an input array with various bits set,
# an Arbiter returns an array with only a single bit set,
# in this case the lowest bit set is retained
#
def DefineArbiter(n):
    T = Bits(n)
    class _Arbiter(Circuit):
        name = 'Arbiter'+str(n)
        IO = ['I', In(T), 'O', Out(T)]
        @classmethod
        def definition(Arb):
            ones = n * [1]
            y = DefineAdd(n)()(Arb.I, array(ones))  # y = x - 1
            def a(y):
                return LUT([0, 1, 0, 0]) # A0 & ~A1
            arb = join(col(a, n))
            arb(Arb.I, y)
            wire(arb.O, Arb.O)
    return _Arbiter

def Arbiter(n, **kwargs):
    return DefineArbiter(n)(**kwargs)

def arbiter(I, **kwargs):
    return Arbiter(len(I), **kwargs)(I)



import magma as m
from .FF import DefineCoreirReg
import mantle


@m.cache_definition
def DefineRegister(n, init=0, has_ce=False, has_reset=False,
                   has_async_reset=False, _type=m.Bits):
    T = _type(n)
    if has_reset and has_async_reset:
        raise ValueError("Cannot have synchronous and asynchronous reset")

    if has_reset or has_ce:
        class Register(m.Circuit):
            name = f"Register_has_ce_{has_ce}_has_reset_{has_reset}_" \
                   f"has_async_reset_{has_async_reset}_" \
                   f"type_{_type.__name__}_n_{n}"
            IO = ["I", m.In(T), "O", m.Out(T)]
            IO += m.ClockInterface(has_ce=has_ce,
                                   has_reset=has_reset,
                                   has_async_reset=has_async_reset)

            @classmethod
            def definition(io):
                reg = DefineCoreirReg(n, init, has_async_reset, _type)(name="value")
                I = io.I
                if has_reset:
                    I = mantle.mux([io.I, m.bits(init, n)], io.RESET)
                if has_ce:
                    I = mantle.mux([reg.O, I], io.CE, name="enable_mux")
                m.wire(I, reg.I)
                m.wire(io.O, reg.O)
        return Register
    else:
        return DefineCoreirReg(n, init, has_async_reset, _type)


def Register(n, init=0, has_ce=False, has_reset=False, has_async_reset=False,
             **kwargs):
    return DefineRegister(n, init, has_ce, has_reset,
                          has_async_reset)(**kwargs)


def register(I, ce=None, reset=None, async_reset=None, **kwargs):
    has_ce = ce is not None
    has_reset = reset is not None
    has_async_reset = async_reset is not None
    reg = Register(len(I),
                   has_ce=has_ce,
                   has_reset=has_reset,
                   has_async_reset=has_async_reset,
                   **kwargs)
    reg(I)
    if has_ce:
        m.wire(ce, reg.CE)
    if has_reset:
        m.wire(reset, reg.RESET)
    if has_async_reset:
        m.wire(reset, reg.ASYNCRESET)



# Programmable logic block functions
from .PLB import *

# Input/output functions
from .IOB import *

# Block RAM functions
from .RAMB import *

# PLL
from .PLL import *

FAMILY = 'ice40'

print('import lattice ice40')



from magma import *
import magma as m
from mantle.coreir.MUX import Mux
import coreir
from bit_vector import BitVector

from .util import DeclareCoreirCircuit

def gen_sim_register(N, init, has_ce, has_reset):
    def sim_register(self, value_store, state_store):
        """
        Adapted from Brennan's SB_DFF simulation in mantle
        """
        cur_clock = value_store.get_value(self.clk)

        if not state_store:
            state_store['prev_clock'] = cur_clock
            state_store['cur_val'] = BitVector(init, num_bits=N) if N is not None else bool(init)

        if has_reset:
            cur_reset = value_store.get_value(self.arst)
        # if s:
        #     cur_s = value_store.get_value(self.S)

        prev_clock = state_store['prev_clock']
        # if not n:
        #     clock_edge = cur_clock and not prev_clock
        # else:
        #     clock_edge = not cur_clock and prev_clock
        clock_edge = cur_clock and not prev_clock

        new_val = state_store['cur_val'].as_bool_list() if N is not None else state_store['cur_val']

        if clock_edge:
            new_val = value_store.get_value(self.I)

        if has_reset and cur_reset:
            new_val = BitVector(init, num_bits=N) if N is not None else bool(init)
        # if s and not sy and cur_s:
        #     new_val = True

        state_store['prev_clock'] = cur_clock
        state_store['cur_val'] = BitVector(new_val, num_bits=N) if N is not None else new_val
        value_store.set_value(self.O, new_val)
    return sim_register


@m.cache_definition
def DefineCoreirReg(width, init=0, has_reset=False, T=Bits):
    if width is None:
        width = 1
    name = "reg_P"  # TODO: Add support for clock interface
    config_args = {"init": coreir.type.BitVector(init, num_bits=width)}
    gen_args = {"width": width}
    T = T(width)
    io = ["I", In(T), "clk", In(Clock), "O", Out(T)]
    methods = []

    def reset(self, condition):
        wire(condition, self.rst)
        return self

    if has_reset:
        io.extend(["arst", In(AsyncReset)])
        name += "R"  # TODO: This assumes ordering of clock parameters
        # methods.append(circuit_type_method("reset", reset))
        # gen_args["has_rst"] = True

    def when(self, condition):
        wire(condition, self.en)
        return self

    # if has_ce:
    #     io.extend(["en", In(Enable)])
    #     name += "E"  # TODO: This assumes ordering of clock parameters
    #     methods.append(circuit_type_method("when", when))
    #     gen_args["has_en"] = True

    # default_kwargs = gen_args.copy()
    default_kwargs = {"init": coreir.type.BitVector(init, num_bits=width)}
    # default_kwargs.update(config_args)

    return DeclareCoreirCircuit(
        name,
        *io,
        stateful=True,
        simulate=gen_sim_register(width, init, False, has_reset),
        circuit_type_methods=methods,
        default_kwargs=default_kwargs,
        coreir_genargs=gen_args,
        coreir_configargs=config_args,
        coreir_name="reg_arst" if has_reset else "reg",
        verilog_name="coreir_" + ("reg_arst" if has_reset else "reg"),
        coreir_lib="coreir"
    )

def define_wrap(type_, type_name, in_type):
    def sim_wrap(self, value_store, state_store):
        input_val = value_store.get_value(getattr(self, "in"))
        value_store.set_value(self.out, input_val)

    return DeclareCircuit(
        f'coreir_wrap{type_name}',
        "in", In(in_type), "out", Out(type_),
        coreir_genargs = {"type": AsyncReset},
        coreir_name="wrap",
        coreir_lib="coreir",
        simulate=sim_wrap
    )

@m.cache_definition
def DefineDFF(init=0, has_ce=False, has_reset=False, has_async_reset=False):
    Reg = DefineCoreirReg(None, init, has_async_reset)
    IO = ["I", In(Bit), "O", Out(Bit)] + ClockInterface(has_ce=has_ce, has_reset=has_reset, has_async_reset=has_async_reset)
    circ = DefineCircuit("DFF_init{}_has_ce{}_has_reset{}_has_async_reset{}".format(init, has_ce, has_reset, has_async_reset),
        *IO)
    reg = Reg()
    wiredefaultclock(circ, reg)
    wireclock(circ, reg)
    I = circ.I
    if has_reset and has_async_reset:
        raise ValueError("Cannot have synchronous and asynchronous reset")
    if has_reset:
        I = Mux()(circ.I, bit(init), circ.RESET)
    if has_ce:
        I = Mux()(reg.O[0], I, circ.CE)
    wire(I, reg.I[0])
    wire(reg.O[0], circ.O)
    EndDefine()
    return circ




def DFF(init=0, has_ce=False, has_reset=False, has_async_reset=False, **kwargs):
    return DefineDFF(init, has_ce, has_reset, has_async_reset)(**kwargs)

FF = DFF



from magma import *
from mantle import And, DefineAdd
from mantle import Mux
from .register import Register
from .counter import _CounterName

__all__ = ['DefineCounterLoad', 'CounterLoad']

#
# Create an n-bit counter with increment and load
#
#   DATA : In(UInt(n)), LOAD : In(Bit), O : Out(UInt(n)), COUT : Out(Bit)
#
def DefineCounterLoad(n, cin=False, cout=True, incr=1, has_ce=False, has_reset=False):

    name = _CounterName(f'CounterLoad{n}', incr, has_ce, has_reset, cin, cout)

    args = []
    args += ['DATA', In(UInt(n))]
    args += ['LOAD', In(Bit)]
    if cin:
        args += ['CIN', In(Bit)]

    args += ["O", Out(UInt(n))]
    if cout:
        args += ["COUT", Out(Bit)]

    args += ClockInterface(has_ce, has_reset)

    Counter = DefineCircuit(name, *args)

    add = DefineAdd(n, cin=cin, cout=cout)()
    mux = Mux(2, n)
    reg = Register(n, has_ce=has_ce, has_reset=has_reset)

    wire( reg.O, add.I0 )
    wire( array(incr, n), add.I1 )

    wire( add.O, mux.I0 )
    wire( Counter.DATA, mux.I1 )
    wire( Counter.LOAD, mux.S )

    reg(mux.O)

    next = False
    if next:
        wire( mux.O, Counter.O )
    else:
        wire( reg.O, Counter.O )

    if cin:
        wire( Counter.CIN, add.CIN )

    if cout:
        wire( add.COUT, Counter.COUT ) # this is fishy because of the LOAD

    wireclock(Counter, reg)

    EndCircuit()

    return Counter

def CounterLoad(n, cin=False, cout=True, incr=1,
        has_ce=False, has_reset=False, **kwargs):
    """Construct a n-bit counter."""
    return DefineCounterLoad(n, cin=cin, cout=cout, incr=incr, 
               has_ce=has_ce, has_reset=has_reset)(**kwargs)





from magma import *
from magma.compatibility import IntegerTypes
from magma.bitutils import int2seq, seq2int
from mantle import FF
from collections import Sequence

__all__  = ['FFs']

__all__ += ['Register', 'DefineRegister', 'register']

__all__ += ['_RegisterName']

#
# Create a column of n FFs initialized to init
#
# Each FF may have a ce, r, and s signal.
#
def FFs(n, init=0, has_ce=False, has_reset=False, has_async_reset=False):
    if isinstance(init, IntegerTypes):
        init = int2seq(init, n)

    def f(y):
        if mantle_target == "coreir":
            return FF(init[y], has_ce=has_ce, has_reset=has_reset, has_async_reset=has_async_reset)
        else:
            if has_async_reset:
                raise NotImplementedError()
            return FF(init[y], has_ce=has_ce, has_reset=has_reset)

    return col(f, n)

## Register module name
def _RegisterName(name, n, init, ce, r):
    name += str(n)
    if ce: name += 'CE'
    if r:  name += 'R'

    if isinstance(init, Sequence):
         init = seq2int(init)
    if init is not 0: name += "_%04X" % init

    return name


def DefineRegister(n, init=0, has_ce=False, has_reset=False, has_async_reset=False, _type=Bits):
    """
    Generate an n-bit register

    Params
    ------
        `_type` - Bits, UInt, or SInt

    Interface
    ---------
        I : In(_type(n)), O : Out(_type(n))
    """
    if _type not in {Bits, UInt, SInt}:
        raise ValueError("Argument _type must be Bits, UInt, or SInt")
    T = _type(n)
    class _Register(Circuit):
        name = _RegisterName('Register', n, init, has_ce, has_reset)
        IO  = ['I', In(T), 'O', Out(T)] + ClockInterface(has_ce=has_ce,has_reset=has_reset,has_async_reset=has_async_reset)
        @classmethod
        def definition(reg):
            ffs = join(FFs(n, init, has_ce, has_reset, has_async_reset))
            wire(reg.I, ffs.I)
            wire(ffs.O, reg.O)
            wireclock(reg, ffs)
            wiredefaultclock(reg, ffs)
    return _Register

def Register(n, init=0, has_ce=False, has_reset=False, has_async_reset=False, **kwargs):
    return DefineRegister(n, init, has_ce, has_reset, has_async_reset)(**kwargs)

def register(I, ce=None, reset=None, async_reset=None, **kwargs):
    has_ce = ce is not None
    has_reset = reset is not None
    has_async_reset = async_reset is not None
    reg = Register(len(I),
                   has_ce=has_ce,
                   has_reset=has_reset,
                   has_async_reset=has_async_reset,
                   **kwargs)
    reg(I)
    if has_ce:
        wire(ce, reg.CE)
    if has_reset:
        wire(reset, reg.RESET)
    if has_async_reset:
        wire(reset, reg.ASYNCRESET)




from magma import *
from magma.backend.verilog import bstr

__all__ = ['SB_PLL', 'SB_PLL40_CORE']

SB_PLL40_CORE = DeclareCircuit('SB_PLL40_CORE',
            "REFERENCECLK", In(Bit),
            "RESETB", In(Bit),
            "BYPASS", In(Bit),
            "PLLOUTCORE", Out(Bit),
            "PLLOUTGLOBAL", Out(Bit))

def SB_PLL( freqout, freqin=12000000 ):

    """Instantiate a PLL"""

    divr, divf, divq, filter = findparams(freqin, freqout)

    params = {}
    params["FEEDBACK_PATH"] = '"SIMPLE"'
    params["PLLOUT_SELECT"] = '"GENCLK"'
    # Reference clock divider (div+1) [0, ..., 15]
    params["DIVR"] = bstr(divr,4)
    # Feedback divider (div+1) [0, ..., 63]
    params["DIVF"] = bstr(divf,7)
    # VCO divider (divq+1) [0, ..., 6]
    params["DIVQ"] = bstr(divq,3)
    params["FILTER_RANGE"] = bstr(filter,3)

    pll = SB_PLL40_CORE(**params)

    wire(1, pll.RESETB)
    wire(0, pll.BYPASS)

    return AnonymousCircuit("I", pll.REFERENCECLK,
                            "O", pll.PLLOUTGLOBAL)

def filterrange(freqin, divr, divf):
    pfd = freqin / (divr + 1)
    vco = pfd * (divf + 1)

    if   pfd <  17: range = 1
    elif pfd <  26: range = 2
    elif pfd <  44: range = 3
    elif pfd <  66: range = 4
    elif pfd < 101: range = 5
    else:           range = 6

    return range

def findparams(freqin, freqout):
    freqin  /= 1000000
    freqout /= 1000000

    bestfreqout = 0.
    for divr in range(16):
        pfd = freqin / (divr + 1)
        if pfd < 10 or pfd > 133:
            continue

        for divf in range(64):
            vco = pfd * (divf + 1)
            if vco < 533 or vco > 1066:
                continue

            # The valid range for DIVQ is 1..6. 
            #
            # This is correctly documented in the ICE Technology Library 
            # Document, but unfortunately incorrectly documented as 0..7 
            # in the iCE40 sysCLOCK PLL Design and Usage Guide.
            #
            for divq in range(1,7):
                fout = vco / (1 << divq)

                # simple mode

                #print('pfd=',pfd, '(divr=',divr+1,')')
                #print('vco=',vco, '(mulf=',divf+1,')')
                #print('freqout=',fout, '(divq=',divq,1<<divq,')')
                if abs(fout - freqout) < abs(bestfreqout - freqout):
                    bestdivr = divr
                    bestdivf = divf
                    bestdivq = divq
                    bestfreqout = fout
                    #print('pfd=',pfd, '(divr=',bestdivr+1,')')
                    #print('vco=',vco, '(mulf=',bestdivf+1,')')
                    #print('bestfreqout=',bestfreqout, '(divq=',bestdivq,1<<bestdivq,')')
    return bestdivr, bestdivf, bestdivq, filterrange(freqin, bestdivr, bestdivf)

#
# from swapforth
#
#  SB_PLL40_CORE #(.FEEDBACK_PATH("SIMPLE"),
#                  .PLLOUT_SELECT("GENCLK"),
#                  .DIVR(4'b0000),
#                  .DIVF(7'd3),
#                  .DIVQ(3'b000),
#                  .FILTER_RANGE(3'b001),
#                 ) uut (
#                         .REFERENCECLK(pclk),
#                         .PLLOUTCORE(clk),
#                         //.PLLOUTGLOBAL(clk),
#                         // .LOCK(D5),
#                         .RESETB(1'b1),
#                         .BYPASS(1'b0)
#                        );


if __name__ == '__main__':
    d = DefineCircuit('main', "CLKIN", In(Bit), "CLKOUT", Out(Bit))
    pll = SB_PLL(25000000, 12000000)
    wire(d.CLKIN, pll.I)
    wire(pll.O, d.CLKOUT)
    EndCircuit()



from magma import *
from .swap import swap
from .halfcleaner import HalfCleaner, ReverseHalfCleaner

__all__  = ['DefineBitonicSorter', 'BitonicSorter', 'bitonicsorter']
__all__ += ['DefineMerger', 'Merger', 'merger']
__all__ += ['DefineSorter', 'Sorter', 'sorter']

# Convert a bitonic sequence into a sorted sequences
#
# HalfCleaner(n) => 2 BitonicSorter(n//2)
#
def DefineBitonicSorter(n):
    assert n in [2, 4, 8, 16]
    T = Bits(n)
    class _BitonicSorter(Circuit):
        name = 'BitonicSorter{}'.format(n)
        IO = ['I', In(T), "O", Out(T)]
        @classmethod
        def definition(io):
            if n == 2:
                wire(swap(io.I), io.O)
            else:
                halfcleaner = HalfCleaner(n)
                bitonic0 = BitonicSorter(n//2)
                bitonic1 = BitonicSorter(n//2)
                bitonic = flat( join(bitonic0, bitonic1), flatargs=['I','O'] )
                s = compose( bitonic, halfcleaner )
                wire( s(io.I), io.O )
    return _BitonicSorter

def BitonicSorter(n):
    return DefineBitonicSorter(n)()

def bitonicsorter(I):
    return BitonicSorter(len(I))(I)


#
# Merge 2 sorted sequencers into a sorted sequence
#
# 2 BitonicSorter(n/2) => ReverseHalfCleaner
#
def DefineMerger(n):
    assert n in [2, 4, 8, 16]
    T = Bits(n)
    class _Merger(Circuit):
        name = 'Merger{}'.format(n)
        IO = ['I', In(T), "O", Out(T)]
        @classmethod
        def definition(io):
            if n == 2:
                wire(swap(io.I), io.O)
            else:
                revhalfcleaner = ReverseHalfCleaner(n)
                bitonic0 = BitonicSorter(n//2)
                bitonic1 = BitonicSorter(n//2)
                bitonic = flat( join(bitonic0, bitonic1), flatargs=['I','O'] )
                s = compose( bitonic, revhalfcleaner )
                wire( s(io.I), io.O )
    return _Merger

def Merger(n):
    return DefineMerger(n)()

def merger(I):
    return Merger(len(I))(I)

#
# Convert an unsorted sequence into a sorted sequence
#
# 2 Sorters(n/2) => Merger(n) 
#
def DefineSorter(n):
    assert n in [2, 4, 8, 16]
    T = Bits(n)
    class _Sorter(Circuit):
        name = 'Sorter{}'.format(n)
        IO = ['I', In(T), "O", Out(T)]
        @classmethod
        def definition(io):
            if n == 2: # Sort 2 element sequences
                wire(swap(io.I), io.O)
            else:
                merger = Merger(n)
                sorter0 = Sorter(n//2) # bot sorter 
                sorter1 = Sorter(n//2) # top sorter
                sorter = flat( join(sorter0, sorter1), flatargs=['I','O'] )
                s = compose( merger, sorter )
                wire( s(io.I), io.O )
    return _Sorter

def Sorter(n):
    return DefineSorter(n)()

def sorter(I):
    return Sorter(len(I))(I)



from magma import *
from .LUT import *
from .arith import DefineSub
from .logic import Not

__all__  = ['DefineEQ', 'EQ']
__all__ += ['DefineNE', 'NE']
__all__ += ['DefineUGE']
__all__ += ['DefineULE']
__all__ += ['DefineUGT']
__all__ += ['DefineULT']
__all__ += ['DefineSGE']
__all__ += ['DefineSLE']
__all__ += ['DefineSGT']
__all__ += ['DefineSLT']

#
# n is the number of luts
# k is the number of bits per lut
# expr goes into each LUT
# fold the luts by wiring O to I0
# extra args can be forked or joined
#
def _Cascade(n, k, expr, cin, forkargs={}):

    assert k <= 3

    def f(y):
        return LUT(expr, k+1)

    cascade = braid( col(f, n), foldargs={"I0":"O"}, forkargs=forkargs )

    wire( cin, cascade.I0 )

    args = []
    if k > 0: args += ["I0", cascade.I1]
    if k > 1: args += ["I1", cascade.I2]
    if k > 2: args += ["I2", cascade.I3]
    args += ["O", cascade.O]

    return AnonymousCircuit(*args)

class EQNone(Circuit):
    IO = ["I0", In(Bit), "I1", In(Bit), "O", Out(Bit)]
    @classmethod
    def definition(io):
        EQ1LUT = ((A0&A1)|(~A0&~A1))
        wire( LUT2(EQ1LUT)(io.I0, io.I1), io.O )

class EQ1(Circuit):
    IO = ["I0", In(Bits(1)), "I1", In(Bits(1)), "O", Out(Bit)]
    @classmethod
    def definition(io):
        wire(EQNone()(io.I0[0], io.I1[0]), io.O)

class EQ2(Circuit):
    T = Bits(2)
    IO = ["I0", In(T), "I1", In(T), "O", Out(Bit)]
    @classmethod
    def definition(io):
        EQ2LUT = ((A0&A1)|(~A0&~A1)) & ((A2&A3)|(~A2&~A3))
        wire( LUT4(EQ2LUT)(io.I0[0], io.I1[0], io.I0[1], io.I1[1]), io.O )

def DefineEQ(n):
    T = Bits(n)
    class _EQ(Circuit):
        name = "EQ{}".format(n)
        IO = ['I0', In(T), 'I1', In(T), "O", Out(Bit)]
        @classmethod
        def definition(io):
             eq = _Cascade(n, 2, A0&((A1&A2)|(~A1&~A2)), 1)
             wire(eq(io.I0, io.I1), io.O)
    return _EQ

def EQ(n, **kwargs):
    if   n == None:
        return EQNone(**kwargs)
    elif n == 1:
        return EQ1(**kwargs)
    elif n == 2:
        return EQ2(**kwargs)
    return DefineEQ(n)(**kwargs)


class NE1(Circuit):
    IO = ["I0", In(Bit), "I1", In(Bit), "O", Out(Bit)]
    @classmethod
    def definition(io):
        NE1LUT = (A0^A1)
        wire( LUT2(NE1LUT)(io.I0, io.I1), io.O )

class NE2(Circuit):
    T = Bits(2)
    IO = ["I0", In(T), "I1", In(T), "O", Out(Bit)]
    @classmethod
    def definition(io):
        NE2LUT = (A0^A1)|(A2^A3)
        wire( LUT4(NE2LUT)(io.I0[0], io.I1[0], io.I0[1], io.I1[1]), io.O )

def DefineNE(n):
    T = Bits(n)
    class _NE(Circuit):
        name = "NE{}".format(n)
        IO = ['I0', In(T), 'I1', In(T), "O", Out(Bit)]
        @classmethod
        def definition(io):
             ne =  _Cascade(n, 2, A0|(A1^A2), 0)
             wire(ne(io.I0, io.I1), io.O)
    return _NE

def NE(n, **kwargs):
    if   n == 1:
        return NE1(**kwargs)
    elif n == 2:
        return NE2(**kwargs)
    return DefineNE(n)(**kwargs)


# unsigned comparisons

def DefineUCMP(opname, reverse, negate, n):
    T = UInt(n)
    class _UCMP(Circuit):
        name = "{}{}".format(opname, n)
        IO = ['I0', In(T), 'I1', In(T), "O", Out(Bit)]
        @classmethod
        def definition(io):
            sub = DefineSub(n, False, True)()
            if not reverse:
                sub(io.I0, io.I1)
            else:
                sub(io.I1, io.I0)
            if not negate:
                wire(sub.COUT, io.O)
            else:
                wire(Not()(sub.COUT), io.O)
    return _UCMP

def DefineUGE(n):
    return DefineUCMP('UGE', False, False, n)

def DefineULE(n):
    return DefineUCMP('ULE', True, False, n)

def DefineULT(n):
    return DefineUCMP('ULT', False, True, n)

def DefineUGT(n):
    return DefineUCMP('UGT', True, True, n)


# signed comparison

def _sge(c_msb, a_msb, b_msb):
    return int((~(a_msb ^ b_msb) & ~c_msb) | (~a_msb & b_msb)) & 1

def _slt(c_msb, a_msb, b_msb):
    return int((~(a_msb ^ b_msb) & c_msb) | (a_msb & ~b_msb)) & 1

def DefineSCMP(opname, op, reverse, n):
    T = SInt(n)
    class _SCMP(Circuit):
        name = "{}{}".format(opname,n)
        IO = ['I0', In(T), 'I1', In(T), "O", Out(Bit)]
        @classmethod
        def definition(io):
            sub = DefineSub(n)()
            cmp = LUT3(op)
            if not reverse:
                wire(cmp( sub(io.I0, io.I1)[-1], io.I0[-1], io.I1[-1] ), io.O)
            else:
                wire(cmp( sub(io.I1, io.I0)[-1], io.I1[-1], io.I0[-1] ), io.O)
    return _SCMP

def DefineSGE(n):
    return DefineSCMP('SGE', _sge, False, n)

def DefineSLE(n):
    return DefineSCMP('SLE', _sge, True, n)

def DefineSLT(n):
    return DefineSCMP('SLT', _slt, False, n)

def DefineSGT(n):
    return DefineSCMP('SGT', _slt, True, n)




from magma.backend.coreir_ import CoreIRBackend
from magma.frontend.coreir_ import CircuitInstanceFromGeneratorWrapper

def Term(cirb: CoreIRBackend, width: int):
    """
    Take in an array of wires and connect it to nothing so that you don't get an unconnected

    :param cirb: The CoreIR backend currently be used
    :param width: The width of the element to absorb
    :return:
    """
    return CircuitInstanceFromGeneratorWrapper(cirb, "coreir", "term", "term_w" + str(width), ["global"],
                                               {"width": width})



from magma import *
from .register import _RegisterName, FFs

__all__ = ['DefineSIPO', 'SIPO']

def DefineSIPO(n, init=0, has_ce=False, has_reset=False):
    """
    Generate Serial-In, Parallel-Out shift register.

    I : In(Bits(n),  O : Bits(n)
    """

    class _SIPO(Circuit):
        name = _RegisterName('SIPO', n, init, has_ce, has_reset)
        IO = ['I', In(Bit), 'O', Out(Bits(n))] + \
                ClockInterface(has_ce,has_reset)
        @classmethod
        def definition(sipo):
            ffs = FFs(n, init=init, has_ce=has_ce, has_reset=has_reset)
            reg = scan(ffs, scanargs={"I":"O"})
            reg(sipo.I)
            wire(reg.O, sipo.O)
            wireclock(sipo, reg)
    return _SIPO

def SIPO(n, init=0, has_ce=False, has_reset=False, **kwargs):
    return DefineSIPO(n, init, has_ce, has_reset)(**kwargs)



from magma import *
from .logic import DefineInvert, Not
from .fulladder import FullAdder

__all__  = ['DefineAdd'] 
__all__ += ['DefineSub']
__all__ += ['DefineNegate']

def _AdderName(basename, n, cin, cout):
    name = f"{basename}{n}"
    if cin: name += '_CIN'
    if cout: name += '_COUT'
    return name

def _AdderArgs(n, cin, cout):
    T = Bits(n)

    args = ["I0", In(T), "I1", In(T)]
    if cin:
        args += ['CIN', In(Bit)]

    args += ["O", Out(T)]
    if cout:
        args += ['COUT', Out(Bit)]

    return args

def DefineAdders(name, n, cin, cout, forkargs=[]):

    def f(y):
        return FullAdder()

    c = braid( col(f, n), foldargs={"CIN":"COUT"}, forkargs=forkargs)

    wire(Adders.I0, c.I0)
    wire(Adders.I1, c.I1)

    wire(c.O, Adders.O)

    if cin:
        wire(Adders.CIN, c.CIN)
    else:
        wire(0, c.CIN)

    if cout:
        wire(c.COUT, Adders.COUT)

#
# create an n-bit Adder from n FullAdders
#
# I0:In(Bits(n)), I1:In(Bits(n)), CIN:In(Bit), O:Out(Bits(n)), COUT:Out(Bit)
#
# if cin, CIN is added to the circuit
# if cout: COUT is added to the circuit
#
def DefineAdd(n, cin=False, cout=False):
    class _Add(Circuit):
        name = _AdderName('Add', n, cin, cout)
        IO = _AdderArgs(n, cin, cout)
        @classmethod
        def definition(io):
            def f(y):
                return FullAdder()

            add = braid( col(f, n), foldargs={"CIN":"COUT"})

            wire(io.I0, add.I0)
            wire(io.I1, add.I1)
            if cin:
                wire(io.CIN, add.CIN)
            else:
                wire(0, add.CIN)
            if cout:
                wire(add.COUT, io.COUT)
            wire(add.O, io.O)
    return _Add
    

def DefineSub(n, cin=False, cout=False):
    class _Sub(Circuit):
        name = _AdderName('Sub', n, cin, cout)
        IO = _AdderArgs(n, cin, cout)
        @classmethod
        def definition(io):
            invert = DefineInvert(n)()
            add =  DefineAdd(n, True, cout)()
            wire(io.I0, add.I0)
            wire(io.I1, invert.I)
            wire(invert.O, add.I1)
            wire(add.O, io.O)
            if cin:
                wire( Not()(io.CIN), add.CIN )
            else:
                wire( 1, add.CIN )
            if cout:
                wire(add.COUT, io.COUT)
    return _Sub
    
    

def DefineNegate(width):
    T = Bits(width)
    class _Negate(Circuit):
        name = 'Negate{}'.format(width)
        IO = ['I', In(T), 'O', Out(T)]
        @classmethod
        def definition(io):
            invert = DefineInvert(width)()
            add =  DefineAdd(width, False, False)()
            wire( add( invert(io.I), (array(1,width))), io.O )
    return _Negate
    



from magma import *
from mantle import Decode

__all__  = ['DefineDecoder', 'Decoder', 'decoder']

def DefineDecoder(n, invert=False):
    """
    n-bit to 2^n decoder.
    """

    class _Decoder(Circuit):
        name = 'Decoder'+str(n)+("Invert" if invert else "")
        IO = ['I', In(Bits(n)), 'O', Out(Bits(1<<n))]
        @classmethod
        def definition(io):
            def decode(y):
                return Decode(y, n, invert)
            dec = fork(col(decode, 1<<n))
            wire(dec(io.I), io.O)
    return _Decoder

def Decoder(n, invert=False, **kwargs):
    return DefineDecoder(n, invert)(**kwargs)

def decoder(I, invert=False, **kwargs):
    return Decoder(len(I), invert=invert, **kwargs)(I)



import magma as m
from mantle import XOr, And, Or

__all__ = [ "FullAdder" ]

class FullAdder(m.Circuit):
    IO = ["I0", m.In(m.Bit), "I1", m.In(m.Bit), "CIN", m.In(m.Bit),
          "O", m.Out(m.Bit), "COUT", m.Out(m.Bit)]

    @classmethod
    def definition(io):
        # Generate the sum
        m.wire( XOr(3)(io.I0, io.I1, io.CIN), io.O)
        # Generate the carry
        m.wire( Or(3)(And(2)(io.I0, io.I1),
                      And(2)(io.I1, io.CIN),
                      And(2)(io.I0, io.CIN)), io.COUT)



from magma import *

__all__  = ['And']
__all__ += ['NAnd']
__all__ += ['Or']
__all__ += ['NOr']
__all__ += ['XOr']
__all__ += ['NXOr']
__all__ += ['Not']
__all__ += ['Buf']

def ins(n=2):
    l = []
    for i in range(n):
        l.append('')
        l.append(In(Bit))
    return l

def DefineAnd(n=2):
    return DeclareCircuit(*(['and', '', Out(Bit)] + ins(n)))

def And(n=2, **kwargs):
    return DefineAnd(n)(**kwargs)


def DefineNAnd(n=2):
    return DeclareCircuit(*(['nand', '', Out(Bit)] + ins(n)))

def NAnd(n=2, **kwargs):
    return DefineNAnd(n)(**kwargs)


def DefineOr(n=2):
    return DeclareCircuit(*(['or', '', Out(Bit)] + ins(n)))

def Or(n=2, **kwargs):
    return DefineOr(n)(**kwargs)


def DefineNOr(n=2):
    return DeclareCircuit(*(['nor', '', Out(Bit)] + ins(n)))

def NOr(n=2, **kwargs):
    return DefineNOr(n)(**kwargs)


def DefineXOr(n=2):
    return DeclareCircuit(*(['xor', '', Out(Bit)] + ins(n)))

def XOr(n=2, **kwargs):
    return DefineXOr(n)(**kwargs)


def DefineNXOr(n=2):
    return DeclareCircuit(*(['nxor', '', Out(Bit)] + ins(n)))

def NXOr(n=2, **kwargs):
    return DefineNXOr(n)(**kwargs)


Not = DeclareCircuit('not', '', Out(Bit), '', In(Bit))
Buf = DeclareCircuit('buf', '', Out(Bit), '', In(Bit))






from ..ice40 import FAMILY, \
    A0, A1, A2, A3, \
    I0, I1, I2, I3, \
    ALL, ANY, PARITY, ZERO, ONE, \
    LUTS_PER_LOGICBLOCK, BITS_PER_LUT, LOG_BITS_PER_LUT

from .LUT import *

from .MUX import *

from .FF import *

from .logic import *
from .decode import *

from .halfadder import *
from .fulladder import *
from .arith import *
from .compare import *

from .memory import Memory

print('import lattice mantle40')



from setuptools import setup

setup(
    name='mantle',
    version='0.1.13',
    url='https://github.com/phanrahan/mantle',
    license='MIT',
    maintainer='Lenny Truong',
    maintainer_email='lenny@cs.stanford.edu',
    description='The magma standard library',
    packages=[
        "mantle",
        "mantle.common",
        "mantle.coreir",
        "mantle.lattice",
        "mantle.lattice.mantle40",
        "mantle.lattice.ice40",
        "mantle.verilog",
        "mantle.primitives",
        "mantle.util",
        "mantle.util.lfsr",
        "mantle.util.sort",
        "mantle.util.compressor",
        "mantle.util.lhca",
    ],

    install_requires=[
        "six",
        "fault>=0.20, <=0.36",
        "magma-lang>=0.1.6, <=0.1.9"
    ],
    python_requires='>=3.6'
)



import magma as m
from mantle import XOr, And

__all__ = [ "HalfAdder" ]

class HalfAdder(m.Circuit):
    IO = ["I", m.In(m.Bit), "CIN", m.In(m.Bit),
          "O", m.Out(m.Bit), "COUT", m.Out(m.Bit)]

    @classmethod
    def definition(io):
        # Generate the sum
        m.wire( XOr(2)(io.I, io.CIN), io.O )
        # Generate the carry
        m.wire( And(2)(io.I, io.CIN), io.COUT )



from __future__ import division


from magma import *
import magma as m
from magma.backend.coreir_ import CoreIRBackend
from magma.frontend.coreir_ import DefineCircuitFromGeneratorWrapper
from .util import DeclareCoreirCircuit
from bit_vector import BitVector

import math


def DefineCoreirMux(width=None):
    N = width
    def simulate(self, value_store, state_store):
        I0 = BitVector(value_store.get_value(self.I0))
        I1 = BitVector(value_store.get_value(self.I1))
        S = BitVector(value_store.get_value(self.S))
        O = I1 if S.as_int() else in0
        value_store.set_value(self.O, O)
    if width is None:
        return DeclareCoreirCircuit("coreir_mux{}".format(N),
            *["I0", In(Bit), "I1", In(Bit), "S", In(Bit),
             "O", Out(Bit)],
            verilog_name="coreir_bitmux",
            coreir_name="mux",
            coreir_lib="corebit",
            simulate=simulate
        )
    else:
        return DeclareCoreirCircuit("coreir_mux{}".format(N),
            *["I0", In(Bits(N)), "I1", In(Bits(N)), "S", In(Bit),
             "O", Out(Bits(N))],
            verilog_name="coreir_mux",
            coreir_name="mux",
            coreir_lib="coreir",
            simulate=simulate,
            coreir_genargs={"width": N}
        )


is_power_of_two = lambda num: num != 0 and ((num & (num - 1)) == 0)


@m.cache_definition
def _declare_muxn(height, width):
    def simulate(self, value_store, state_store):
        sel = BitVector(value_store.get_value(self.I.sel))
        out = BitVector(value_store.get_value(self.I.data[sel.as_int]))
        value_store.set_value(self.O, out)
    return DeclareCoreirCircuit(f"coreir_commonlib_mux{height}x{width}",
        *["I", In(Tuple(data=Array(height, Bits(width)),
                         sel=Bits(m.bitutils.clog2(height)))),
          "O", Out(Bits(width))],
        coreir_name="muxn",
        coreir_lib="commonlib",
        simulate=simulate,
        coreir_genargs={"width": width, "N": height}
    )


@m.cache_definition
def DefineMux(height=2, width=None, T=None):
    if T is not None:
        assert width is None, "Can only specify width **or** T"
        # Sanitize names for verilog by removing parens
        # TODO: Make this a reuseable feature
        suffix = str(T).replace("(", "").replace(")", "").replace(",", "_")
        T = T
    else:
        suffix = f"{width}"
        if width is None:
            T = Bit
        else:
            T = Bits(width)

    io = []
    for i in range(height):
        io += ["I{}".format(i), In(T)]
    if height == 2:
        select_type = Bit
    else:
        select_type = Bits(m.bitutils.clog2(height))
    io += ['S', In(select_type)]
    io += ['O', Out(T)]

    class _Mux(Circuit):
        name = "Mux{}x{}".format(height, suffix)
        IO = io
        @classmethod
        def definition(interface):
            if T is not None and not (isinstance(T, m.BitKind) or isinstance(T, m.ArrayKind) and isinstance(T.T, m.BitKind)):
                if isinstance(T, m.TupleKind):
                    raise NotImplementedError()
                else:
                    assert isinstance(T, m.ArrayKind), f"Expected array or type type, got {T}, type is {type(T)}"
                    for i in range(len(T)):
                        Is = [getattr(interface, f"I{j}")[i] for j in range(height)]
                        interface.O[i] <= DefineMux(height, T=type(Is[0]))()(*Is, interface.S)
            else:
                if T is None and width is None or isinstance(T, m.BitKind):
                    mux = _declare_muxn(height, 1)()
                else:
                    mux = _declare_muxn(height, width if T is None else T.size())()
                for i in range(height):
                    if T is None and width is None or isinstance(T, m.BitKind):
                        m.wire(getattr(interface, f"I{i}"), mux.I.data[i][0])
                    else:
                        m.wire(getattr(interface, f"I{i}"), mux.I.data[i])
                if height == 2:
                    m.wire(interface.S, mux.I.sel[0])
                else:
                    m.wire(interface.S, mux.I.sel)
                if T is None and width is None or isinstance(T, m.BitKind):
                    wire(mux.O[0], interface.O)
                else:
                    wire(mux.O, interface.O)
    return _Mux


def Mux(height=2, width=None, T=None, **kwargs):
    return DefineMux(height, width, T=T)(**kwargs)

Mux2 = DefineMux(2)
Mux4 = DefineMux(4)
Mux8 = DefineMux(8)
Mux16 = DefineMux(16)

def DefineCommonlibMuxN(cirb: CoreIRBackend, N: int, width: int):
    """
    Get a Mux that handles any number of inputs

    :param N: The number of inputs to select between
    :param width: The number of bits per input
    :return: A circuit with the following ports:
    I : {
        data: In(Array(N, Array(width, Bit)))
        sel: In(Array(getRAMAddrWidth(N), Bit))
    }
    out : Out(Array(width, Bit))
    Note: even though this isn't a RAM, the AddrWidth computation is the same.
    """
    name = "CommonlibMuxN_n{}_w{}".format(str(N), str(width))
    return DefineCircuitFromGeneratorWrapper(cirb, "commonlib", "muxn",
                                             name, ["mantle", "coreir", "global"],
                                             {"N": N, "width": width})

def CommonlibMuxN(cirb: CoreIRBackend, N: int, width: int):
    return DefineCommonlibMuxN(cirb, N, width)()



from magma import *
from mantle import Mux
from .register import _RegisterName, Register, FFs

__all__ = ['DefinePISO', 'PISO']

def DefinePISO(n, init=0, has_ce=False, has_reset=False):
    """
    Generate Parallel-In, Serial-Out shift register.

    SI : In(Bit), PI : Out(Bits(n)), LOAD : In(Bit), O : Out(Bit)
    """

    T = Bits(n)
    class _PISO(Circuit):
        name = _RegisterName('PISO', n, init, has_ce, has_reset)
        IO = ['SI', In(Bit), 'PI', In(T), 'LOAD', In(Bit),
              'O', Out(Bit)] + ClockInterface(has_ce,has_reset)
        @classmethod
        def definition(piso):
            def mux2(y):
                return curry(Mux(2), prefix='I')

            mux = braid(col(mux2, n), forkargs=['S'])
            reg = Register(n, init, has_ce=has_ce, has_reset=has_reset)

            #si = array(*[piso.SI] + [reg.O[i] for i in range(n-1)])
            si = concat(array(piso.SI),reg.O[0:n-1])
            mux(si, piso.PI, piso.LOAD)
            reg(mux)
            wire(reg.O[n-1], piso.O)
            wireclock(piso, reg)

    return _PISO

def PISO(n, init=0, has_ce=False, has_reset=False, **kwargs):
    return DefinePISO(n, init, has_ce, has_reset)(**kwargs)




from magma import Circuit, Bit, Bits, In, Out, bits, wire, map_
from magma.bitutils import log2
from mantle import Mux2

__all__  = ['DefineShift', 'Shift']
__all__  = ['DefineLSL', 'LSL']
__all__ += ['DefineLSR', 'LSR']
__all__ += ['DefineASR', 'ASR']

__all__ += ['DefineRotate', 'Rotate']
__all__ += ['DefineROL', 'ROL']
__all__ += ['DefineROR', 'ROR']

def DefineShiftK(n, k, op):
    assert k < n
    T = Bits(n)
    class _ShiftK(Circuit):
        name = f'{op.upper()}{n}_{k}'
        IO = ['I', In(T), 'S', In(Bit), "O", Out(T)]
        @classmethod
        def definition(io):
            Is = [io.I[i] for i in range(n)]
            muxes = map_(Mux2, n)
            for i in range(n):
                if   op == 'lsl':
                    shifti = i - k
                    I = bits([Is[i], Is[shifti] if shifti >= 0 else 0])
                elif op == 'lsr':
                    shifti = i + k
                    I = bits([Is[i], Is[shifti] if shifti <  n else 0])
                elif op == 'asr':
                    shifti = i + k
                    I = bits([Is[i], Is[shifti] if shifti <  n else Is[n-1]])
                else:
                    assert False
                muxes[i]( I, io.S )
            for i in range(n):
                Is[i] = muxes[i].O
            wire(bits(Is), io.O)
    return _ShiftK

def ShiftK(n, k, op):
    return DefineShiftK(n, k, op)()

def DefineShift(n, op):
    assert n in [2, 4, 8, 16]
    logn = log2(n)
    T = Bits(n)
    class _Shift(Circuit):
        name = f'{op.upper()}{n}'
        IO = ['I', In(T), 'S', In(Bits(logn)), "O", Out(T)]
        @classmethod
        def definition(io):
            I = io.I
            for k in range(logn):
                 I = ShiftK(n, 1<<k, op)(I, io.S[k])
            wire(I, io.O)
    return _Shift

def Shift(n, op):
    return DefineShift(n, op)()


def DefineLSL(n):
    return DefineShift(n, 'lsl')

def LSL(n):
    return Shift(n, 'lsl')

def DefineLSR(n):
    return DefineShift(n, 'lsr')

def LSR(n):
    return Shift(n, 'lsr')

def DefineASR(n):
    return DefineShift(n, 'asr')

def ASR(n):
    return Shift(n, 'asr')



def DefineRotateK(n, k, op):
    assert k < n
    T = Bits(n)
    class _RotateK(Circuit):
        name = f'{op.upper()}{n}_{k}'
        IO = ['I', In(T), 'S', In(Bit), "O", Out(T)]
        @classmethod
        def definition(io):
            Is = [io.I[i] for i in range(n)]
            muxes = map_(Mux2, n)
            for i in range(n):
                if  op == 'rol':
                    rotatei = (i - k + n) % n
                    I = bits([Is[i], Is[rotatei]])
                elif op == 'ror':
                    rotatei = (i + k) % n
                    I = bits([Is[i], Is[rotatei]])
                else:
                    assert False
                muxes[i]( I, io.S )
            for i in range(n):
                Is[i] = muxes[i].O
            wire(bits(Is), io.O)
    return _RotateK

def RotateK(n, k, op):
    return DefineRotateK(n, k, op)()

def DefineRotate(n, op):
    assert n in [2, 4, 8, 16]
    logn = log2(n)
    T = Bits(n)
    class _Rotate(Circuit):
        name = f'{op.upper()}{n}'
        IO = ['I', In(T), 'S', In(Bits(logn)), "O", Out(T)]
        @classmethod
        def definition(io):
            I = io.I
            for k in range(logn):
                 I = RotateK(n, 1<<k, op)(I, io.S[k])
            wire(I, io.O)
    return _Rotate

def Rotate(n, op):
    return DefineRotate(n, op)()

def DefineROL(n):
    return DefineRotate(n, 'rol')

def ROL(n):
    return Rotate(n, 'rol')

def DefineROR(n):
    return DefineRotate(n, 'ror')

def ROR(n):
    return Rotate(n, 'ror')



from functools import wraps

import magma as m
from magma.bitutils import seq2int
from mantle import And, NAnd, Or, NOr, XOr, NXOr, LSL, LSR, Not, Invert, UDiv, SDiv, UMod, SMod, Mul
from mantle import ASR
from mantle import EQ, NE, ULT, ULE, UGT, UGE, SLT, SLE, SGT, SGE
from mantle import Mux
from .arith import Add, Sub, Negate

__all__ = []


def get_length(value):
    if isinstance(value, m._BitType):
        return None
    elif isinstance(value, m.ArrayType):
        return len(value)
    elif isinstance(value, int):
        return value.bit_length()
    else:
        raise NotImplementedError(f"Cannot get_length of {type(value)}")

def check_widths(width, args):
    for arg in args:
        arg_len = get_length(arg)
        # Allow integers with widths less than the width (they will be
        # promoted)
        if arg_len != width \
                and not ((isinstance(arg, int) and arg_len is 1 and width is None) or \
                         (isinstance(arg, int) and arg_len < width)):
            raise ValueError(
                f"All arguments should have the same length: {args}")


def check_operator_args(fn):
    @wraps(fn)
    def wrapped(*args, **kwargs):
        if len(args) < 2:
            raise RuntimeError(
                f"{fn.__name__} requires at least 2 arguments")
        width = get_length(args[0])
        check_widths(width, args)
        T = type(args[0])
        if not (all(issubclass(type(x).__class__, T.__class__) for x in args) or (
                all(issubclass(T.__class__, type(x).__class__) for x in args))):
            raise TypeError(
                "Currently Arguments to operators must be of the same type")
        return fn(width, *args, **kwargs)
    return wrapped


operators = {}


def _pass_closure_vars_as_args(*closure_args):
    def _wrapped(fn):
        @wraps(fn)
        def _wrapped_inner(*args, **kwargs):
            return fn(*closure_args, *args, **kwargs)
        return _wrapped_inner
    return _wrapped


def export(fn):
    global __all__
    __all__ += [fn.__name__]
    return fn


def preserve_type(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        retval = fn(*args, **kwargs)
        T = type(args[0])
        if isinstance(T, m.UIntKind):
            return m.uint(retval)
        elif isinstance(T, m.SIntKind):
            return m.sint(retval)
        return retval
    return wrapper


for _operator_name, _Circuit in (
    ("and_", And),
    ("nand", NAnd),
    ("or_", Or),
    ("nor", NOr),
    ("xor", XOr),
    ("nxor", NXOr),
    ("add", Add),
    ("sub", Sub),
    ("mul", Mul),
    ("udiv", UDiv),
    ("sdiv", SDiv),
    ("umod", UMod),
    ("smod", SMod)
):
    __all__ += [_operator_name]

    # Because Python uses dynamic scoping, the closures will use the
    # last value of _Circuit rather than capturing the lexical value.
    # Hacky workaround is to pass _Circuit as an argument to a
    # decorator so the "lexical" value is captured.
    @preserve_type
    @check_operator_args
    @_pass_closure_vars_as_args(_Circuit, _operator_name)
    def operator(circuit, name, width, *args, **kwargs):
        for arg in args:
            if not isinstance(arg, (m.Type, int)):
                raise TypeError("Operators only defined on magma types and Python ints")
        if name == "le" and args[0].isinput :
            assert len(args) == 2, "Expected two arguments for assignment (<=)"
            if not isinstance(args[1], int) and not args[1].isoutput():
                raise TypeError("Can only assign an output to an input")
        else:
            for arg in args:
                if not isinstance(arg, int) and not arg.isoutput():
                    raise TypeError("Non-assigment operators are only defined on output types")

        if name == "le" and args[0].isinput():
            # We could just call wire here, but we dispatch to the default
            # magma implementation so we only have to maintain the logic in one
            # place
            m.Type.__le__(args[0], args[1])
        else:
            if name in ["add", "sub", "mul", "udiv", "sdiv", 'umod', 'smod']:
                # These don't have a height
                if len(args) > 2:
                    raise Exception(f"{name} operator expects 2 arguments")
                return circuit(width, **kwargs)(*args)
            else:
                return circuit(len(args), width, **kwargs)(*args)
    operator.__name__ = _operator_name
    operator.__qualname__ = _operator_name
    operators[_operator_name] = operator
    exec(f"{_operator_name} = operator")


@export
@preserve_type
def lsl(I0, I1, **kwargs):
    width = get_length(I0)
    return LSL(width, **kwargs)(I0, I1)


@export
@preserve_type
def lsr(I0, I1, **kwargs):
    width = get_length(I0)
    return LSR(width, **kwargs)(I0, I1)


@export
@preserve_type
def asr(I0, I1, **kwargs):
    width = get_length(I0)
    return ASR(width, **kwargs)(I0, I1)


@export
@preserve_type
def not_(arg, **kwargs):
    return Not(**kwargs)(arg)


@export
@preserve_type
def invert(arg, **kwargs):
    width = get_length(arg)
    if width is None:
        return Not(**kwargs)(arg)
    else:
        return Invert(width, **kwargs)(arg)


@export
@preserve_type
def neg(arg, **kwargs):
    if isinstance(arg, int):
        return -arg
    return Negate(get_length(arg), **kwargs)(arg)


bitwise_ops = [
    ("__and__", and_),
    ("__or__", or_),
    ("__xor__", xor),
    ("__invert__", invert),
    ("__lshift__", lsl),
    ("__rshift__", lsr),
]

for method, op in bitwise_ops:
    if op not in (lsl, lsr):
        setattr(m.BitType, method, op)
    setattr(m.BitsType, method, op)

arithmetic_ops = [
    ("__neg__", neg),
    ("__add__", add),
    ("__sub__", sub),
    ("__mul__", mul),
]


@export
@check_operator_args
def lt(width, I0, I1, **kwargs):
    if isinstance(I0, m.SIntType):
        return SLT(width, **kwargs)(I0, I1)
    else:
        return ULT(width, **kwargs)(I0, I1)


@export
@check_operator_args
def le(width, I0, I1, **kwargs):
    if I0.isinput():
        return m.Type.__le__(I0, I1)
    if isinstance(I0, m.SIntType):
        return SLE(width, **kwargs)(I0, I1)
    else:
        return ULE(width, **kwargs)(I0, I1)


@export
@check_operator_args
def gt(width, I0, I1, **kwargs):
    if isinstance(I0, m.SIntType):
        return SGT(width, **kwargs)(I0, I1)
    else:
        return UGT(width, **kwargs)(I0, I1)


@export
@check_operator_args
def ge(width, I0, I1, **kwargs):
    if isinstance(I0, m.SIntType):
        return SGE(width, **kwargs)(I0, I1)
    else:
        return UGE(width, **kwargs)(I0, I1)


@export
@check_operator_args
def eq(width, I0, I1, **kwargs):
    return EQ(width, **kwargs)(I0, I1)


@export
@check_operator_args
def ne(width, I0, I1, **kwargs):
    return NE(width, **kwargs)(I0, I1)


relational_ops = [
    ("__lt__", lt),
    ("__le__", le),
    ("__gt__", gt),
    ("__ge__", ge),
]

for method, op in arithmetic_ops + relational_ops:
    setattr(m.SIntType, method, op)
    setattr(m.UIntType, method, op)

m.SIntType.__truediv__ = sdiv
m.UIntType.__truediv__ = udiv

m.SIntType.__mod__ = smod
m.UIntType.__mod__ = umod

for type_ in (m._BitType, m.ArrayType):
    setattr(type_, "__eq__", eq)
    setattr(type_, "__neq__", ne)


@export
@preserve_type
def mux(I, S, **kwargs):
    if isinstance(S, int):
        return I[S]
    elif S.const():
        return I[seq2int(S.bits())]
    return Mux(len(I), T=type(I[0]), **kwargs)(*I, S)


orig_get_item = m.ArrayType.__getitem__


def dynamic_mux_select(self, S):
    if isinstance(S, m.Type):
        if isinstance(self.T, m._BitKind):
            length = None
        else:
            length = len(self.T)
        height = len(self)
        inputs = self.ts[:]
        if m.mantle_target == "ice40" and height not in [2, 4, 8, 16]:
            if height > 16:
                raise NotImplementedError()
            orig_height = height
            height = 2 ** m.bitutils.clog2(height)
            if not isinstance(inputs[0], m._BitType):
                raise NotImplementedError(type(inputs[0]))
            inputs.extend([m.bit(0) for _ in range(height - orig_height)])

        return Mux(height, length)(*inputs, S)
    return orig_get_item(self, S)


setattr(m.ArrayType, "__getitem__", dynamic_mux_select)



import magma as m
from mantle import FF, And

#
# debounce 
#  a - input signal
#  slow sets the delay
#
def debounce(a, slow):
    ffa = FF(has_ce=True)
    ffa(a, ce=slow)

    ffb = FF(has_ce=True)
    ffb(ffa.O, ce=slow)

    # return the value of a if it is stable for slow duration
    return And(2)(ffa.O, ffb.O)



import magma as m
from mantle import DefineAdd, DefineSub, DefineNegate

__all__ = [ "Add", "Sub", "Negate", "UDiv", "SDiv", "UMod", "SMod", "Mul"]


def Add(n, cin=False, cout=False, **kwargs):
    return DefineAdd(n, cin=cin, cout=cout)(**kwargs)


def Sub(n, cin=False, cout=False, **kwargs):
    return DefineSub(n, cin=cin, cout=cout)(**kwargs)


def Negate(width, **kwargs):
    return DefineNegate(width)(**kwargs)


def Mul(width, **kwargs):
    if m.mantle_target != "coreir":
        raise NotImplementedError(
            "Mul not implemented for mantle target {m.mantle_target}")
    # Workaround import error for other targets by delaying import until after
    # if guard
    from mantle import DefineMul
    return DefineMul(width)(**kwargs)


def UDiv(width, **kwargs):
    if m.mantle_target != "coreir":
        raise NotImplementedError(
            "Division not implemented for mantle target {m.mantle_target}")
    # Workaround import error for other targets by delaying import until after
    # if guard
    from mantle import DefineUDiv
    return DefineUDiv(width)(**kwargs)


def SDiv(width, **kwargs):
    if m.mantle_target != "coreir":
        raise NotImplementedError(
            "Division not implemented for mantle target {m.mantle_target}")
    # Workaround import error for other targets by delaying import until after
    # if guard
    from mantle import DefineSDiv
    return DefineSDiv(width)(**kwargs)


def UMod(width, **kwargs):
    if m.mantle_target != "coreir":
        raise NotImplementedError(
            "Modulo not implemented for mantle target {m.mantle_target}")
    # Workaround import error for other targets by delaying import until after
    # if guard
    from mantle import DefineUMod
    return DefineUMod(width)(**kwargs)


def SMod(width, **kwargs):
    if m.mantle_target != "coreir":
        raise NotImplementedError(
            "Modulo not implemented for mantle target {m.mantle_target}")
    # Workaround import error for other targets by delaying import until after
    # if guard
    from mantle import DefineSMod
    return DefineSMod(width)(**kwargs)
